import fnmatch
import json
import logging
from datetime import datetime

import boto3
import pendulum
from pydantic import BaseModel, HttpUrl, ValidationError
from tabulate import tabulate

from .common import session_maker, validate_response

logger = logging.getLogger(__file__)

KST = pendulum.timezone("Asia/Seoul")


################################################################
# Model for S3 Database
################################################################
class ValueModel(BaseModel):
    _filename_field = "url"

    url: HttpUrl
    method: str = None
    created_at: datetime = datetime.now().astimezone(KST)
    finished_at: datetime = None
    modified_count: int = 1


################################################################
# S3 Common Helpers
################################################################
def list_objects(
    bucket: str,
    prefix: str = None,
    pattern: str = None,
    session: boto3.Session = None,
    session_opts: dict = None,
):
    # correct
    prefix = prefix if prefix else ""
    pattern = "*" + pattern if pattern else pattern

    # get session
    session = session if session else session_maker(session_opts=session_opts)
    client = session.client("s3")

    # get list of objects
    contents = []
    paginator = client.get_paginator("list_objects_v2")
    for response in paginator.paginate(Bucket=bucket, Prefix=prefix):
        for content in response.get("Contents", []):
            if not pattern or fnmatch.fnmatch(content["Key"], pattern):
                contents.append(content)

    return contents


def delete_objects(
    bucket: str,
    prefix: str,
    pattern: str = None,
    confirm: bool = True,
    session: boto3.Session = None,
    session_opts: dict = None,
):
    # correct
    prefix = prefix if prefix else ""

    # get session
    session = session if session else session_maker(session_opts=session_opts)
    client = session.client("s3")

    # list objects to delete
    contents = list_objects(bucket=bucket, prefix=prefix, pattern=pattern, session=session)
    n_objects = len(contents)

    if n_objects == 0:
        logger.warning("no objects to delete. exit!")
        return

    if confirm:
        print(f"Find {n_objects} Objects to Delete.\n" + tabulate(contents))
        answer = input("Are You Sure? (yes or no) ")
        if answer != "yes":
            print("Exit!")
            return

    # delete
    if len(contents) > 0:
        # delete objects
        delete = {"Objects": [{"Key": content["Key"]} for content in contents]}
        response = client.delete_objects(Bucket=bucket, Delete=delete)
        return response["Deleted"]

    raise FileNotFoundError(f"no pattern '{pattern}' in prefix '{bucket}/{prefix}'!")


################################################################
# S3 as a Database
################################################################
class Database:
    def __init__(
        self,
        bucket: str,
        prefix: str,
        Model: BaseModel = ValueModel,
        session: boto3.Session = None,
        session_opts: dict = None,
    ):
        # props
        self.bucket = bucket
        self.prefix = prefix
        self.Model = Model

        # session
        self.session = session if session else session_maker(session_opts=session_opts)
        self.client = self.session.client("s3")
        self.resource = self.session.resource("s3")
        self.Bucket = self.resource.Bucket(self.bucket)

    def keys(self, patterns: list = ["*"], unfinished=None):
        objects = self._list_objects(patterns=patterns)
        if unfinished is None:
            return [self._trim_key(obj.key) for obj in objects]

        # filter unfinished, finished
        keys = [obj.key for obj in objects]
        objects = [self._deserialize(body) for obj in objects for body in obj.get()["Body"]]
        if unfinished:
            return [
                self._trim_key(key)
                for key, obj in zip(keys, objects)
                if obj is not None and obj.created_at is not None
            ]
        else:
            return [
                self._trim_key(key) for key, obj in zip(keys, objects) if obj is not None and obj.created_at is None
            ]

    def clean(self):
        """Delete Wrong Schema Keys"""
        objects = self._list_objects()
        for obj in objects:
            for body in obj.get()["Body"]:
                try:
                    value = self._deserialize(body)
                except Exception as ex:
                    logger.error(ex)
                    raise ex
                if value is None:
                    key = self._trim_key(obj.key)
                    logger.warning(f"object {key} has wrong schema - removed!")
                    print(key)
                    self.delete(key=key)

    def _list_objects(self, patterns: list = ["*"]):
        return [
            obj
            for obj in self.Bucket.objects.filter(Prefix=self.prefix)
            if any([fnmatch.fnmatch(obj.key, "/".join([self.prefix, pat])) for pat in patterns])
        ]

    def set(
        self,
        key: str,
        url: HttpUrl,
        method: str = None,
        finished_at: datetime = None,
        overwrite: bool = True,
    ):
        if finished_at and not isinstance(finished_at, datetime):
            finished_at = datetime.fromisoformat(finished_at)
        value = ValueModel(url=url, method=method, finished_at=finished_at)
        return self._set(key=key, value=value, overwrite=overwrite)

    def _set(self, key: str, value: ValueModel, overwrite=True):
        current = self._get(key)
        if current is not None:
            if not overwrite:
                raise FileExistsError(f"object {self._generate_full_key(key)} exists!")
            value.modified_count = current.modified_count + 1
            value.created_at = current.created_at

        params = {
            "Bucket": self.bucket,
            "Key": self._generate_full_key(key),
            "Body": self._serialize(value),
        }

        try:
            response = self.client.put_object(**params)
            validate_response(response)
            return response
        except Exception as ex:
            logger.error(ex)
            raise ex

    def get(
        self,
        key: str,
        if_modified_since: datetime = None,
        if_unmodified_since: datetime = None,
    ):
        value = self._get(key=key, if_modified_since=if_modified_since, if_unmodified_since=if_unmodified_since)
        if value is None:
            return
        return value.dict()

    def _get(
        self,
        key: str,
        if_modified_since: datetime = None,
        if_unmodified_since: datetime = None,
    ):
        params = {
            "Bucket": self.bucket,
            "Key": self._generate_full_key(key),
        }
        if if_modified_since:
            if not isinstance(if_modified_since, datetime):
                if_modified_since = datetime.fromisoformat(if_modified_since)
            params.update({"IfModifiedSince": if_modified_since})
        if if_unmodified_since:
            if not isinstance(if_unmodified_since, datetime):
                if_unmodified_since = datetime.fromisoformat(if_unmodified_since)
            params.update({"IfUnmodifiedSince": if_unmodified_since})

        try:
            response = self.client.get_object(**params)
            validate_response(response)
        except self.client.exceptions.NoSuchKey:
            return

        for body in response["Body"]:
            return self._deserialize(body)

    def delete(self, key: str):
        Object = {
            "Bucket": self.bucket,
            "Key": self._generate_full_key(key),
        }
        response = self.client.delete_object(**Object)
        validate_response(response, success_codes=[200, 204])
        return response

    def _generate_full_key(self, key):
        return "/".join([self.prefix, key.strip("/")])

    def _trim_key(self, full_key):
        return full_key.replace(f"{self.prefix}/", "", 1)

    def _serialize(self, value):
        return value.json()

    def _deserialize(self, body):
        try:
            _body = json.loads(body)
            return self.Model(**_body)
        except ValidationError:
            logger.warning(f"skip! deserialize failed, {body}")
            return
        except Exception as ex:
            logger.error(ex)
            raise ex



import boto3

from .. import docker

SESSION_OPTS = {
    "aws_access_key_id": None,
    "aws_secret_access_key": None,
    "aws_session_token": None,
    "region_name": None,
    "botocore_session": None,
    "profile_name": None,
}


################################################################
# Helpers
################################################################
# create client for SecretsManager
def session_maker(
    profile_name=None,
    aws_access_key_id=None,
    aws_secret_access_key=None,
    region_name="ap-northeast-2",
    load_docker_secret=True,
    session_opts=None,
):
    """Create boto3 secretsmanager client.

    Priority:
        1. profile_name
        2. aws_access_key_id & secret_access_key
        3. docker secret (/run/secret)
    """

    # session configuration
    session_opts = session_opts or dict()
    if load_docker_secret:
        docker.load_secrets()
    if region_name is not None:
        session_opts.update({"region_name": region_name})
    if aws_access_key_id is not None:
        if aws_secret_access_key is not None:
            session_opts.update(
                {
                    "aws_access_key_id": aws_access_key_id,
                    "aws_secret_access_key": aws_secret_access_key,
                }
            )
    if profile_name is not None:
        session_opts = {"profile_name": profile_name}

    # return client
    return boto3.session.Session(**session_opts)


def validate_response(response, success_codes=[200]):
    meta = response["ResponseMetadata"]
    if meta["HTTPStatusCode"] not in success_codes:
        raise ReferenceError(f"status code {meta['HTTPStatusCode']}, {str(meta)}")



import fnmatch
import json
import logging
from typing import Union

import boto3
from rich import print

from .common import session_maker

logger = logging.getLogger()

PS_PREFIX = "clutter"


################################################################
# HELPERS
################################################################
# exception
class ClutterAWSException(Exception):
    pass


################################################################
# FUNCTIONS
################################################################
# list parameters
def list_parameters(
    patterns: Union[str, list] = None,
    session: boto3.Session = None,
) -> list:
    """List Parameters in AWS Parameter Store.

    Args:
        session (boto3.Session, optional): create new session if None. Defaults to None.

    Returns:
        list: list of parameters
    """
    # correct args
    patterns = patterns or "*"
    if patterns:
        patterns = patterns if isinstance(patterns, (tuple, list)) else [patterns]

    # get client
    session = session or session_maker()
    client = session.client("ssm")

    parameter_filters = [{"Key": "Name", "Option": "BeginsWith", "Values": [f"/{PS_PREFIX}"]}]
    resp = client.describe_parameters(ParameterFilters=parameter_filters)

    for param in resp["Parameters"]:
        _, name = param["Name"].strip("/").split("/")
        if not any([fnmatch.fnmatch(name, f"*{pat.strip('*')}*") for pat in patterns]):
            continue
        print(f"[[bold]{name}[/bold]]")
        if param.get("Description"):
            print(f" DESCRIPTION: {param['Description']}")
        print(f""" SSM NAME: '{param["Name"]}'""")
        print(" VALUES:")
        values = get_parameters(name)
        if isinstance(values, dict):
            for k, v in values.items():
                print(f"  - {k}: {v}")
        else:
            print(f"  - {values}")


# set parameters
def set_parameters(
    name: str,
    parameters: dict,
    description: str = None,
    tags: dict = None,
    overwrite: bool = False,
    session: boto3.Session = None,
) -> None:
    """Set New Parameters.

    Args:
        name (str): Name.
        parameters (dict): i.e. {"host": "example.com", "port": 80, "user": "eugene"}
        description (str, optional): i.e. "This is Description.". Defaults to None.
        tags (dict, optional): i.e. {"creator": "eugene"}. Defaults to None.
        overwrite (bool, optional): overwrite exist parameters if True. Defaults to False.
        session (boto3.Session, optional): create new session if None. Defaults to None.

    Raises:
        ClutterAWSException
    """
    # get client
    session = session or session_maker()
    client = session.client("ssm")

    opts = {
        "Type": "SecureString",
        "Tier": "Standard",
    }
    if description:
        opts.update({"Description": description})
    if tags:
        Tags = []
        for k, v in tags.items():
            tags.append({"Key": k, "Value": v})
        opts.update({"Tags": Tags})

    resp = client.put_parameter(
        Name=f"/{PS_PREFIX}/{name}",
        Value=json.dumps(parameters) if isinstance(parameters, dict) else parameters,
        Overwrite=overwrite,
        **opts,
    )
    if resp["ResponseMetadata"]["HTTPStatusCode"] != 200:
        raise ClutterAWSException(str(resp))


# get parameters
def get_parameters(name: str, session: boto3.Session = None) -> Union[dict, str]:
    """Get Parameters from AWS Parameter Store.

    Args:
        name (str): name of parameters
        session (boto3.Session, optional): create new session if None. Defaults to None.

    Returns:
        Union[dict, str]: parameters.
    """
    session = session or session_maker()
    client = session.client("ssm")

    Name = f"/{PS_PREFIX}/{name}"
    resp = client.get_parameter(Name=Name, WithDecryption=True)
    values = resp["Parameter"]["Value"]
    try:
        return json.loads(values)
    except json.JSONDecodeError as exc:
        return values



import logging
import os
from pathlib import Path

logger = logging.getLogger(__file__)

SECRETS_DIR = "/run/secrets"
SECRETS_DELIM = "__"


def load_secrets(
    secrets_dir: str = SECRETS_DIR,
    secrets_delim: str = SECRETS_DELIM,
):
    """
    Read docker secrets and set environment variables.

    Examples
    --------
    >>> # 'AWS_ACCESS_KEY_ID__eugene_personal' is in '/run/secrets' folder
    >>> os.getenv("AWS_ACCESS_KEY_ID")
    >>> load_secrets()
    >>> os.getenv("AWS_ACCESS_KEY_ID")
    <eugene's access key id>
    """

    if not os.path.exists(secrets_dir):
        logger.debug(f"No docker secrets - '{secrets_dir}' not exists.")
        return

    SECRET_DIR = Path(SECRETS_DIR)
    for env, fp in [(x.name.rsplit(secrets_delim, 1)[0], x) for x in SECRET_DIR.glob("*") if x.is_file()]:
        try:
            with open(fp, "r") as f:
                lines = f.readlines()
                for line in lines:
                    if line.strip() != "":
                        os.environ[env] = line.rstrip()
                        break
            logger.debug(f"Env. '{env}' is loaded from docker secrets.")
        except Exception as ex:
            logger.warning(f"Loading env. '{env}' is failed! - {ex}")



import json
import time
from typing import Dict, Union

import boto3
from pydantic import BaseModel

import logging
from .common import session_maker, validate_response


logger = logging.getLogger(__file__)

################################################################
# Models
################################################################
class Message(BaseModel):
    Id: str = None
    MessageBody: str = None
    MessageAttributes: dict = None
    MessageSystemAttributes: dict = None
    MessageDuplicationId: str = None
    MessageGroupId: str = None


class ReturnedMessage(BaseModel):
    MessageId: str = None
    ReceiptHandle: str = None
    MD5OfMessageBody: str = None
    Body: str = None
    MessageAttributes: dict = None


################################################################
# Queue Hadlers
################################################################
# list queues
def list_queues(prefix=None, session=None, session_opts=None):
    session = session if session else session_maker(session_opts=session_opts)
    client = session.client("sqs")

    paginate_opts = {
        "QueueNamePrefix": prefix,
        "PaginationConfig": {},
    }
    paginate_opts = {k: v for k, v in paginate_opts.items() if v is not None}

    queue_urls = []
    paginator = client.get_paginator("list_queues")
    for response in paginator.paginate(**paginate_opts):
        for queue_url in response.get("QueueUrls", []):
            queue_urls.append(queue_url)

    return queue_urls


# create queue
def create_queue(
    queue_name: str,
    delay_seconds: int = 0,
    visibility_timeout: int = 60,
    dlq_after_received: int = 10,
    wait_for_queue_to_ready_sec: int = -1,
    session: boto3.Session = None,
    session_opts: dict = None,
):
    """
    [NOTE]
    자동 생성한 DLQ는 queue와 동일한 속성을 가짐.

    Arguments
    ---------
    dql_after_received: int
        consumer가 받은 횟수가 이 값을 초과하면 DLQ로 보냄. 0보다 작게 설정하면 DLQ 사용하지 않음.
    wait_for_queue_to_ready_sec: int
        queue가 생성된 것을 확인한 후에 response를 반환. (queue 생성은 주문 후 30초 이상 걸리기도 함.)
    """
    session = session if session else session_maker(session_opts=session_opts)
    client = session.client("sqs")

    _queues = list_queues(prefix=queue_name)
    for _queue in _queues:
        if _queue.rsplit("/", 1)[-1] == queue_name:
            logger.warning(f"queue '{queue_name}' already exists!")
            return

    redrive_policy = {}
    if dlq_after_received > 0:
        dlq_name = f"{queue_name}-dlq"
        response = create_queue(
            queue_name=dlq_name,
            delay_seconds=delay_seconds,
            visibility_timeout=visibility_timeout,
            dlq_after_received=-1,
            wait_for_queue_to_ready_sec=60,
        )
        dlq_url = get_queue_url(queue_name=dlq_name, session=session)
        response = client.get_queue_attributes(QueueUrl=dlq_url, AttributeNames=["QueueArn"])
        _ = validate_response(response)
        dlq_arn = response["Attributes"]["QueueArn"]
        redrive_policy = {
            "deadLetterTargetArn": dlq_arn,
            "maxReceiveCount": str(dlq_after_received),
        }

    attributes = {
        "DelaySeconds": str(delay_seconds),
        "VisibilityTimeout": str(visibility_timeout),
    }
    if redrive_policy:
        attributes.update({"RedrivePolicy": json.dumps(redrive_policy)})

    response = client.create_queue(
        QueueName=queue_name,
        Attributes=attributes,
    )
    _ = validate_response(response)

    if wait_for_queue_to_ready_sec > 0:
        t0 = time.time()
        logger.warning(
            f"wait for the new queue '{queue_name}' to be ready... (maximum {wait_for_queue_to_ready_sec} senconds)"
        )
        for _ in range(wait_for_queue_to_ready_sec):
            time.sleep(1)
            try:
                _ = get_queue(queue_name=queue_name, session=session)
                logger.warning(f"queue '{queue_name}' is created! ({time.time() - t0:.3f} sec.)")
                break
            except KeyError:
                pass
            except Exception as ex:
                logger.error(ex)
                raise ex

    return response


# get queue
def get_queue(
    queue_name: str,
    create_if_not_exists: bool = False,
    create_if_not_exists_timeout: int = 60,
    queue_delay_seconds: int = 0,
    queue_visibility_timeout: int = 900,
    dlq_after_received: int = 10,
    session: boto3.Session = None,
    session_opts: dict = None,
):
    session = session if session else session_maker(session_opts=session_opts)
    resource = session.resource("sqs")

    for queue in resource.queues.all():
        if queue.url.rsplit("/", 1)[-1] == queue_name:
            return queue

    if create_if_not_exists:
        response = create_queue(
            queue_name=queue_name,
            delay_seconds=queue_delay_seconds,
            visibility_timeout=queue_visibility_timeout,
            dlq_after_received=dlq_after_received,
            wait_for_queue_to_ready_sec=create_if_not_exists_timeout,
            session=session,
        )
        _ = validate_response(response)
        return get_queue(queue_name=queue_name, session=session)

    raise KeyError(f"queue '{queue_name}' not found!")


# get queue
def set_queue_attributes(
    queue_name: str,
    delay_seconds: int = None,
    maximum_message_size: int = None,
    message_retention_period: int = None,
    receive_message_wait_time_seconds: int = None,
    dlq_queue_name: str = None,
    dlq_max_recevice_count: int = None,
    session: boto3.Session = None,
    session_opts: dict = None,
):
    session = session if session else session_maker(session_opts=session_opts)
    client = session.client("sqs")

    attributes = {
        "DelaySeconds": delay_seconds,
        "MaximumMessageSize": maximum_message_size,
        "MessageRetentionPeriod": message_retention_period,
        "ReceviceMessageWaitTimeSeconds": receive_message_wait_time_seconds,
    }
    attributes = {k: str(v) for k, v in attributes.items() if v is not None}
    if dlq_queue_name is not None or dlq_max_recevice_count is not None:
        dlq_queue_url = get_queue_url(queue_name=dlq_queue_name, session=session)
        response = client.get_queue_attributes(QueueUrl=dlq_queue_url, AttributeNames=["QueueArn"])
        _ = validate_response(response)
        dlq_arn = response["Attributes"]["QueueArn"]
        redrive_policy = {
            "deadLetterTargetArn": dlq_arn,
            "maxReceiveCount": dlq_max_recevice_count,
        }
        redrive_policy = {k: str(v) for k, v in redrive_policy.items() if v is not None}

    response = client.set_queue_attributes(
        QueueUrl=get_queue_url(queue_name),
        Attributes=attributes,
    )
    _ = validate_response(response)

    return response


# get queue url
def get_queue_url(
    queue_name: str,
    session: boto3.Session = None,
    session_opts: dict = None,
):
    session = session if session else session_maker(session_opts=session_opts)
    client = session.client("sqs")

    response = client.get_queue_url(QueueName=queue_name)
    _ = validate_response(response)

    return response["QueueUrl"]


# delete queue
def delete_queue(
    queue_name: str,
    delete_dlq: bool = False,
    session: boto3.Session = None,
    session_opts: dict = None,
):
    session = session if session else session_maker(session_opts=session_opts)
    client = session.client("sqs")

    if delete_dlq:
        queue = get_queue(queue_name=queue_name, create_if_not_exists=False, session=session)
        redrive_policy = queue.attributes.get("RedrivePolicy")
        if redrive_policy:
            redrive_policy = json.loads(redrive_policy)
            _queue_name = redrive_policy["deadLetterTargetArn"].rsplit(":", 1)[-1]
            dlq_sources = list_dead_letter_source_queues(queue_name=_queue_name, session=session)
            if len(dlq_sources) > 1:
                logger.warning(f"other queue uses '{_queue_name}' as DLQ, we'll not delete it!")
            delete_queue(queue_name=_queue_name)

    queue_url = get_queue_url(queue_name=queue_name)
    response = client.delete_queue(QueueUrl=queue_url)
    _ = validate_response(response)

    return response


# delete queue
def purge_queue(
    queue_name: str,
    session: boto3.Session = None,
    session_opts: dict = None,
):
    session = session if session else session_maker(session_opts=session_opts)
    client = session.client("sqs")

    queue_url = get_queue_url(queue_name=queue_name)
    response = client.purge_queue(QueueUrl=queue_url)
    _ = validate_response(response)

    return response


# list dead letter source queues
def list_dead_letter_source_queues(queue_name, session=None, session_opts=None):
    session = session if session else session_maker(session_opts=session_opts)
    client = session.client("sqs")

    queue_url = get_queue_url(queue_name=queue_name, session=session)

    paginate_opts = {
        "QueueUrl": queue_url,
        "PaginationConfig": {},
    }
    paginate_opts = {k: v for k, v in paginate_opts.items() if v is not None}

    queue_urls = []
    paginator = client.get_paginator("list_dead_letter_source_queues")
    for response in paginator.paginate(**paginate_opts):
        for queue_url in response.get("queueUrls", []):
            queue_urls.append(queue_url)

    return queue_urls


################################################################
# Message Handler
################################################################
# Queue
class Queue:
    def __init__(
        self,
        queue_name: str,
        auto_create_queue: bool = False,
        session: boto3.Session = None,
        session_opts: dict = None,
    ):
        # correct args
        session_opts = session_opts if session_opts else {}

        # props
        self.session = session if session else session_maker(session_opts)
        self.client = self.session.client("sqs")
        self.resource = self.session.resource("sqs")
        self.queue = get_queue(queue_name=queue_name, session=self.session, create_if_not_exists=auto_create_queue)
        self.queue_url = get_queue_url(queue_name, session=self.session)

    def send_messages(self, messages: list):
        if not isinstance(messages, (tuple, list)):
            raise ValueError("input 'messages' should be a list!")

        # request validator
        messages = [msg if isinstance(msg, Message) else Message(**msg) for msg in messages]
        entries = [msg.dict(exclude_none=True) for msg in messages]

        # send messages
        response = self.queue.send_messages(Entries=entries)
        _ = validate_response(response)
        if "Successful" in response:
            n_send = len(messages)
            n_succeed = len(response["Successful"])
            if n_succeed != n_send:
                logger.warning(f"send {n_send} messages, only {n_succeed} messages are succeed!")
            return response["Successful"]

        raise ReferenceError("send messages FAILED!")

    def receive_messages(self, message_attribute_names=["All"], max_number_of_messages=10, wait_time_seconds=1):
        response = self.client.receive_message(
            QueueUrl=self.queue_url,
            MessageAttributeNames=message_attribute_names,
            MaxNumberOfMessages=max_number_of_messages,
            WaitTimeSeconds=wait_time_seconds,
        )
        _ = validate_response(response)
        if "Messages" in response:
            for msg in response["Messages"]:
                print(msg)
            return [ReturnedMessage(**msg) for msg in response["Messages"]]
        return []

    def delete_message(self, receipt_handle_or_message: Union[str, ReturnedMessage]):
        if isinstance(receipt_handle_or_message, ReturnedMessage):
            receipt_handle_or_message = receipt_handle_or_message.ReceiptHandle
        response = self.client.delete_message(QueueUrl=self.queue_url, ReceiptHandle=receipt_handle_or_message)
        _ = validate_response(response)
        return response

    def purge_queue(self):
        response = self.client.purge_queue(QueueUrl=self.queue_url)
        _ = validate_response(response)
        return response



import logging
import os

import click
from click_loglevel import LogLevel
from InquirerPy import prompt, inquirer
from tabulate import tabulate

from . import aws

logger = logging.getLogger(__file__)


@click.group()
def dasida():
    pass


################################################################
# AWS SecretsManager
################################################################
@dasida.group()
def secretsmanager():
    pass


@secretsmanager.command()
@click.argument("patterns", default="*")
@click.option("--profile", default=None)
def list(patterns, profile):
    aws.secretsmanager.list_secrets(patterns=patterns, profile_name=profile)


@secretsmanager.command()
@click.argument("secret_name")
@click.option("--profile", default=None)
def get(secret_name, profile):
    secrets = aws.secretsmanager.get_secrets(secret_name=secret_name, profile_name=profile)
    if secrets:
        print(f"Secret '{secret_name}':")
        for k, v in secrets.items():
            print(f" - {k}: {v}")


@secretsmanager.command()
@click.argument("secret_name")
@click.option("--profile", default=None)
def delete(secret_name, profile):
    _ = aws.secretsmanager.delete_secrets(secret_name=secret_name, profile_name=profile)
    print(f"Secret '{secret_name}' is deleted!")


@secretsmanager.command()
@click.argument("secret_name")
@click.option("--profile", default=None)
def create(secret_name, profile):
    if aws.secretsmanager.get_secrets(secret_name=secret_name, profile_name=profile):
        logger.error("Secret '{}' is already exist!".format(secret_name))
        return

    secrets = dict()
    while True:
        key = inquirer.text(message="Key:").execute()
        value = inquirer.text(message="Value:").execute()
        secrets.update({key: value})
        proceed = inquirer.confirm(message="Add New Key-Value?").execute()
        if not proceed:
            break

    print(f"\nAdd new secret '{secret_name}' as follows?:")
    for k, v in secrets.items():
        print(f" - {k}: {v}")

    proceed = inquirer.confirm(message="Confirm?", default=False).execute()
    if proceed:
        _ = aws.secretsmanager.create_secrets(secret_name=secret_name, secrets=secrets, profile_name=profile)
    else:
        logger.error("Abort!")


@secretsmanager.command()
@click.argument("secret_name")
@click.option("--profile", default=None)
def update(secret_name, profile):
    secrets = aws.secretsmanager.get_secrets(secret_name=secret_name, profile_name=profile)
    if not secrets:
        msg = f"Secret name '{secret_name}' not found!"
        raise KeyError(msg)

    new_secrets = dict()
    for k, v in secrets.items():
        key = inquirer.text(message="Key:", default=k).execute()
        value = inquirer.text(message="Value:", default=v).execute()
        if value:
            new_secrets.update({key: value})

    while True:
        proceed = inquirer.confirm(message="Add New Key-Value?").execute()
        if not proceed:
            break
        key = inquirer.text(message="Key:").execute()
        value = inquirer.text(message="Value:").execute()
        new_secrets.update({key: value})

    print(f"\nUpdate secret '{secret_name}' as follows?:")
    for k, v in new_secrets.items():
        print(f" - {k}: {v}")

    proceed = inquirer.confirm(message="Confirm?", default=False).execute()
    if proceed:
        _ = aws.secretsmanager.update_secrets(secret_name=secret_name, secrets=new_secrets, profile_name=profile)
    else:
        logger.error("Abort!")


################################################################
# AWS S3
################################################################
@dasida.group()
def s3():
    pass


# list objects
@s3.command()
@click.option("-b", "--bucket", type=str, required=True, help="S3 Bucket Name.")
@click.option("--prefix", type=str, default=None, help="Object Prefix. Example: 'dataset_a/daily'")
@click.option("--pattern", type=str, default=None, help="Pattern. Example: '2022-*-*-file.csv'")
@click.option("--profile", type=str, default=None, help="AWS Profile Name.")
@click.option("--log-level", type=LogLevel(), default=logging.INFO, help="AWS Profile Name.")
def list_objects(bucket, prefix, pattern, profile, log_level):
    # set log level
    logger.setLevel(level=log_level)

    # create session
    session = aws.common.session_maker(profile_name=profile)

    # get results
    contents = aws.s3.list_objects(bucket, prefix, pattern, session=session)

    # print results
    path = "/".join([x for x in [bucket, prefix, pattern] if x is not None])
    print(f"ls {path}\n" + tabulate(contents))


# delete objects
@s3.command()
@click.option("-b", "--bucket", type=str, required=True, help="S3 Bucket Name.")
@click.option("--prefix", type=str, default=None, help="Object Prefix. Example: 'dataset_a/daily'")
@click.option("--pattern", type=str, default=None, help="Pattern. Example: '2022-*-*-file.csv'")
@click.option("--profile", type=str, default=None, help="AWS Profile Name.")
@click.option("--log-level", type=LogLevel(), default=logging.INFO, help="AWS Profile Name.")
def delete_objects(bucket, prefix, pattern, profile, log_level):
    # set log level
    logger.setLevel(level=log_level)

    # create session
    session = aws.common.session_maker(profile_name=profile)

    # get results
    response = aws.s3.delete_objects(bucket, prefix, pattern, session=session)
    if response is None:
        return

    # print results
    path = "/".join([x for x in [bucket, prefix, pattern] if x is not None])

    # logging
    print(f"total {len(response)} objects are deleted from {path}!")


################################################################
# AWS SQS
################################################################
@dasida.group()
def sqs():
    pass


@sqs.command()
@click.option("--prefix", type=str, default=None)
@click.option("--profile", type=str, default=None, help="AWS Profile Name.")
@click.option("--log-level", type=LogLevel(), default=logging.INFO, help="Log Level.")
def list_queues(prefix, profile, log_level):
    # set log level
    logger.setLevel(level=log_level)

    # create session
    session = aws.common.session_maker(profile_name=profile)

    # get queue urls
    queue_urls = aws.sqs.list_queues(prefix=prefix, session=session)
    print("List of Queues.\n" + tabulate([(q.rsplit("/", 1)[-1], q) for q in sorted(queue_urls)]))


@sqs.command()
@click.argument("queue-name", type=str, default=None)
@click.option("--delete-dlq", type=bool, default=True)
@click.option("--profile", type=str, default=None, help="AWS Profile Name.")
@click.option("--log-level", type=LogLevel(), default=logging.INFO, help="Log Level.")
def delete_queue(queue_name, delete_dlq, profile, log_level):
    # set log level
    logger.setLevel(level=log_level)

    # create session
    session = aws.common.session_maker(profile_name=profile)

    # get queue urls
    try:
        _ = aws.sqs.delete_queue(queue_name=queue_name, delete_dlq=delete_dlq, session=session)
    except KeyError:
        print(f"No Queue '{queue_name}' was Found. exit!")
        return
    except Exception as ex:
        logger.error(ex)
        return

    print(f"Queue '{queue_name}' is Deleted.")


@sqs.command()
@click.argument("queue-name", type=str, default=None)
@click.option("--profile", type=str, default=None, help="AWS Profile Name.")
@click.option("--log-level", type=LogLevel(), default=logging.INFO, help="Log Level.")
def purge_queue(queue_name, profile, log_level):
    # set log level
    logger.setLevel(level=log_level)

    # create session
    session = aws.common.session_maker(profile_name=profile)

    # get queue urls
    try:
        _ = aws.sqs.purge_queue(queue_name=queue_name, session=session)
    except KeyError:
        print(f"No Queue '{queue_name}' was Found. exit!")
        return
    except Exception as ex:
        logger.error(ex)
        return

    print(f"Queue '{queue_name}' is Purged.")


@sqs.command()
@click.argument("queue-name", type=str, default=None)
@click.option("--profile", type=str, default=None, help="AWS Profile Name.")
@click.option("--log-level", type=LogLevel(), default=logging.INFO, help="Log Level.")
def get_queue_url(queue_name, profile, log_level):
    # set log level
    logger.setLevel(level=log_level)

    # create session
    session = aws.common.session_maker(profile_name=profile)

    # get queue urls
    try:
        queue_url = aws.sqs.get_queue_url(queue_name=queue_name, session=session)
    except KeyError:
        print(f"No Queue '{queue_name}' was Found. exit!")
        return
    except Exception as ex:
        logger.error(ex)
        return

    print(f"QueueUrl: {queue_url}")


@sqs.command()
@click.argument("queue-name", type=str, default=None)
@click.option("--delay-seconds", type=int, default=0)
@click.option("--visibility-timeout", type=int, default=60)
@click.option("--dlq_after_received", type=int, default=10, help="Not Create DLQ if This Value is Minus")
@click.option("--profile", type=str, default=None, help="AWS Profile Name.")
@click.option("--log-level", type=LogLevel(), default=logging.INFO, help="Log Level.")
def create_queue(queue_name, delay_seconds, visibility_timeout, dlq_after_received, profile, log_level):
    # set log level
    logger.setLevel(level=log_level)

    # create session
    session = aws.common.session_maker(profile_name=profile)

    # get queue urls
    try:
        response = aws.sqs.create_queue(
            queue_name=queue_name,
            delay_seconds=delay_seconds,
            visibility_timeout=visibility_timeout,
            dlq_after_received=dlq_after_received,
            wait_for_queue_to_ready_sec=60,
            session=session,
        )
    except Exception as ex:
        print(f"Create Queue '{queue_name}' Failed. Exit!")
        logger.error(ex)
        return

    if response is None:
        return
    print(f"Queue '{queue_name}' Created. (QueueUrl: {response['QueueUrl']})")



import ast
import base64
import fnmatch
import json
import logging
from typing import Union, List

import boto3
from botocore.exceptions import ClientError
from rich import print

from .common import session_maker

logger = logging.getLogger(__file__)


################################################################
# HELPERS
################################################################
# _get_secrets
def _get_secrets(client, secret_name):
    # get secrets
    try:
        get_secret_value_response = client.get_secret_value(SecretId=secret_name)
    except ClientError as e:
        code = e.response["Error"]["Code"]
        description = ERROR_DESCRIPTIONS.get(code, "unknown").replace("\n", " ")
        if code == "ResourceNotFoundException":
            logger.info(f"{code}: secret '{secret_name}' not found! return None.")
            return None
        logger.error(f"{code}: {description}")
        raise e
    else:
        if "SecretString" in get_secret_value_response:
            secrets = get_secret_value_response["SecretString"]
            try:
                secrets = json.loads(secrets)
            except json.JSONDecodeError:
                try:
                    secrets = ast.literal_eval(secrets)
                except SyntaxError as ex:
                    pass
            except Exception as ex:
                logger.error(f"[DASIDA] Cannot Decode JSON, {secrets}")
        else:
            secrets = base64.b64decode(get_secret_value_response["SecretBinary"])

    return secrets


################################################################
# FUNCTIONS
################################################################
# list secrets
def list_secrets(
    patterns: Union[str, list] = "*",
    profile_name: str = None,
    aws_access_key_id: str = None,
    aws_secret_access_key: str = None,
    region_name: str = "ap-northeast-2",
    load_docker_secret: bool = True,
    session: boto3.Session = None,
):
    """
    (TODO)
      - filter tags
    """
    # correct args
    if patterns:
        patterns = patterns if isinstance(patterns, (tuple, list)) else [patterns]

    # get client
    session = session or session_maker(
        profile_name=profile_name,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=region_name,
        load_docker_secret=load_docker_secret,
    )
    client = session.client("secretsmanager")

    # get secrets
    opts = {"MaxResults": 100}
    secrets = []
    while True:
        response = client.list_secrets(**opts)
        secrets += response.get("SecretList", [])
        next_token = response.get("NextToken")
        if next_token is None:
            break
        opts.update({"NextToken": next_token})

    for secret in secrets:
        secret_name = secret["Name"]
        if not any([fnmatch.fnmatch(secret_name, f"*{pat.strip('*')}*") for pat in patterns]):
            continue
        try:
            body = _get_secrets(client, secret_name)
        except Exception as ex:
            print(f"[FAIL] {secret_name}")
        print(f"[[bold]{secret['Name']}[/bold]]")
        if secret.get("Description"):
            print(f" DESCRIPTION: {secret['Description']}")
        print(" VALUES:")
        if isinstance(body, dict):
            for k, v in body.items():
                print(f"  - {k}: {v}")
        else:
            print(f"  - {body}")


# get secrets
def get_secrets(
    secret_name: str,
    profile_name: str = None,
    aws_access_key_id: str = None,
    aws_secret_access_key: str = None,
    region_name: str = "ap-northeast-2",
    load_docker_secret: bool = True,
    session: boto3.Session = None,
):
    """Get secrets from AWS SecretsManager.

    Example
    -------
    >>> conf = get_secrets("some/secrets")
    """

    # get client
    session = session or session_maker(
        profile_name=profile_name,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=region_name,
        load_docker_secret=load_docker_secret,
    )
    client = session.client("secretsmanager")

    return _get_secrets(client, secret_name)


# create secrets
def create_secrets(
    secret_name: str,
    secrets: Union[str, dict],
    description: str = None,
    tags: List[dict] = None,
    profile_name: str = None,
    aws_access_key_id: str = None,
    aws_secret_access_key: str = None,
    region_name: str = "ap-northeast-2",
    load_docker_secret: bool = True,
    session: boto3.Session = None,
):
    """Create secrets from AWS SecretsManager.

    Example
    -------
    >>> create_secrets("some/secrets")
    """

    # get client
    session = session or session_maker(
        profile_name=profile_name,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=region_name,
        load_docker_secret=load_docker_secret,
    )
    client = session.client("secretsmanager")

    if isinstance(secrets, dict):
        secrets = json.dumps(secrets)

    opts = {"Name": secret_name, "SecretString": secrets}
    if description:
        opts.update({"Description": description})
    if tags:
        opts.update({"Tags": tags})

    response = client.create_secret(**opts)

    return response


# update secrets
def update_secrets(
    secret_name: str,
    secrets: Union[str, dict] = None,
    description: str = None,
    tags: List[dict] = None,
    profile_name: str = None,
    aws_access_key_id: str = None,
    aws_secret_access_key: str = None,
    region_name: str = "ap-northeast-2",
    load_docker_secret: bool = True,
    session: boto3.Session = None,
):
    """Update secrets from AWS SecretsManager.

    Example
    -------
    >>> update_secrets("some/secrets")
    """

    # get client
    session = session or session_maker(
        profile_name=profile_name,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=region_name,
        load_docker_secret=load_docker_secret,
    )
    client = session.client("secretsmanager")

    opts = {"SecretId": secret_name}
    new_opts = dict()
    if secrets:
        if isinstance(secrets, dict):
            secrets = json.dumps(secrets)
        new_opts.update({"SecretString": secrets})
    if description:
        new_opts.update({"Description": description})
    if tags:
        new_opts.update({"Tags": tags})

    if not new_opts:
        logger.warning("Nothing to update. Exit!")

    opts.update(new_opts)
    response = client.update_secret(**opts)

    return response


# put secrets
def put_secrets(
    secret_name: str,
    secrets: Union[str, dict] = None,
    profile_name: str = None,
    aws_access_key_id: str = None,
    aws_secret_access_key: str = None,
    region_name: str = "ap-northeast-2",
    load_docker_secret: bool = True,
    session: boto3.Session = None,
):
    """Update secrets from AWS SecretsManager.

    Example
    -------
    >>> put_secrets("some/secrets")
    """

    # get client
    session = session or session_maker(
        profile_name=profile_name,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=region_name,
        load_docker_secret=load_docker_secret,
    )
    client = session.client("secretsmanager")

    if isinstance(secrets, dict):
        current_secrets = get_secrets(secret_name, session=session)
        current_secrets.update(secrets)
        secrets = json.dumps(current_secrets)

    opts = {"SecretId": secret_name, "SecretString": secrets}
    response = client.put_secret_value(**opts)

    return response


# delete secrets
def delete_secrets(
    secret_name: str,
    recovery_window_in_days: int = None,
    force_delete_without_recovery: bool = None,
    profile_name: str = None,
    aws_access_key_id: str = None,
    aws_secret_access_key: str = None,
    region_name: str = "ap-northeast-2",
    load_docker_secret: bool = True,
    session: boto3.Session = None,
):
    """Create secrets from AWS SecretsManager.

    Example
    -------
    >>> create_secrets("some/secrets")
    """

    # get client
    session = session or session_maker(
        profile_name=profile_name,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=region_name,
        load_docker_secret=load_docker_secret,
    )
    client = session.client("secretsmanager")

    opts = {"SecretId": secret_name}
    if recovery_window_in_days:
        opts.update({"RecoveryWindowInDasy": recovery_window_in_days})
    if force_delete_without_recovery:
        opts.update({"ForceDeleteWithoutRecovery": force_delete_without_recovery})

    response = client.delete_secret(**opts)

    return response


# errors
ERROR_DESCRIPTIONS = {
    "DecryptionFailureException": """\
    Secrets Manager can't decrypt the protected secret text using the provided KMS key.""",
    "InternalServiceErrorException": """\
    An error occurred on the server side.""",
    "InvalidParameterException": """\
    You provided an invalid value for a parameter.""",
    "InvalidRequestException": """\
    You provided a parameter value that is not valid for the current state of the resource.""",
    "ResourceNotFoundException": """\
    We can't find the resource that you asked for.""",
}
